(* Repair Thunderbird encrypted mails
   ==================================
   Use this program to repair Thunderbird mail files that will no more be decrypted
   by Enigmail since version 2.0.6 as a result of a missing
   Modification Detection Code (MDC)

    J. Rathlev, D-24222 Schwentinental (kontakt(a)rathlev-home.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   March 2019
   *)

unit MailRepairMain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, Vcl.Buttons, Vcl.ExtCtrls,
  Vcl.ComCtrls;

const
  ProgName = 'Repair Thunderbird encrypted mails';
  Vers = ' - Vers. 1.0';
  CopRgt = ' 2010-2019, Dr. J. Rathlev';
  EmailAdr = 'kontakt(a)rathlev-home.de';

  GpgDecCommand = 'gpg --output "%1:s" --status-fd 1 -d "%0:s"';
  GpgEncCommand = 'gpg --output "%2:s" --armor -e -r %0:s "%1:s"';
  sGpgPlaintext = '[GNUPG:] PLAINTEXT ';
  AttPrefix = 'X-Content-PGP-Universal-Saved-';

type
  TBlock = record
    BlkStart,BlkEnd : integer;
    end;

  TfrmMailDecrypt = class(TForm)
    leTbMailFile: TLabeledEdit;
    btTbFile: TBitBtn;
    OpenDialog: TOpenDialog;
    paResult: TPanel;
    laCount: TLabel;
    Label1: TLabel;
    btnDecrypt: TBitBtn;
    laEncCount: TLabel;
    StatusBar: TStatusBar;
    procedure FormCreate(Sender: TObject);
    procedure btTbFileClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure btnDecryptClick(Sender: TObject);
    procedure leTbMailFileChange(Sender: TObject);
    procedure FormShow(Sender: TObject);
  private
    { Private-Deklarationen }
    ProgVersName,
    ProgVersDate,
    UserPath,AppPath,
    MailFile   : string;
    MailText   : TStringList;
    EncIndex   : array of TBlock;
    procedure AnalyzeFile;
  public
    { Public-Deklarationen }
  end;

var
  frmMailDecrypt: TfrmMailDecrypt;

implementation

{$R *.dfm}

uses System.StrUtils, GnuGetText, InitProg, WinUtils, FileUtils,
  StringUtils, WinExecute, WinApiUtils;

procedure TfrmMailDecrypt.FormCreate(Sender: TObject);
begin
  TranslateComponent(self);
  InitPaths(AppPath,UserPath);
  InitVersion(ProgName,Vers,CopRgt,3,3,ProgVersName,ProgVersDate);
  MailFile:=''; EncIndex:=nil;
  MailText:=TStringList.Create;
  end;

procedure TfrmMailDecrypt.FormDestroy(Sender: TObject);
begin
  EncIndex:=nil;
  MailText.Free;
  end;

procedure TfrmMailDecrypt.FormShow(Sender: TObject);
begin
//  MailFile:='e:\Jrgen\test\Mail\testb';
//  leTbMailFile.Text:=MailFile;
//  AnalyzeFile;
  end;

procedure TfrmMailDecrypt.leTbMailFileChange(Sender: TObject);
begin
  MailText.Clear; laCount.Caption:=''; laEncCount.Caption:='';
  end;

procedure TfrmMailDecrypt.btTbFileClick(Sender: TObject);
begin
  with OpenDialog do begin
    if length(MailFile)>0 then InitialDir:=ExtractFilePath(MailFile)
    else InitialDir:=UserPath;
    Filename:=ExtractFilename(MailFile);
    Filter:=_('Thunderbird mail files|*.*');
    Title:=_('Select Thunderbird mail file');
    if Execute then begin
      leTbMailFile.Text:=Filename;
      MailFile:=Filename;
      AnalyzeFile;
      end;
    end;
  end;

procedure TfrmMailDecrypt.AnalyzeFile;
var
  sl      : string;
  i,nm,ne : integer;
  blk     : boolean;
begin
  if FileExists(MailFile) then begin
    laCount.Caption:=_('Analyzing file');
    laEncCount.Caption:='';
    Application.ProcessMessages;
    MailText.LoadFromFile(MailFile);
    nm:=0; ne:=0; blk:=false;
    SetLength(EncIndex,MailText.Count);
    for i:=0 to MailText.Count-1 do begin
      sl:=Trim(MailText[i]);
      if AnsiStartsText('From - ',sl) then inc(nm)
      else if AnsiStartsText('-----BEGIN PGP MESSAGE-----',sl) then begin
        blk:=true;
        with EncIndex[ne] do BlkStart:=i;
        end
      else if blk and AnsiStartsText('-----END PGP MESSAGE-----',sl) then begin
        blk:=false;
        EncIndex[ne].BlkEnd:=i;
        inc(ne);
        end;
      end;
    SetLength(EncIndex,ne);
    laCount.Caption:=Format(_('Total number of mails: %u'),[nm]);
    laEncCount.Caption:=Format(_('Number of encrypted sections: %u'),[ne]);
    end
  else ErrorDialog(_('File not found!'));
  end;

procedure TfrmMailDecrypt.btnDecryptClick(Sender: TObject);
var
  i,j,nl,n,k          : integer;
  hr                  : HResult;
  oMail,oBlock,iBlock : TextFile;
  PgpID,
  sName,sbe,sbd,s     : string;
  StdOut              : TStringList;
begin
  if MailText.Count=0 then begin
    if IsEmptyStr(leTbMailFile.Text) then btTbFileClick(Sender)
    else begin
      MailFile:=leTbMailFile.Text;
      if FileExists(MailFile) then AnalyzeFile
      else begin
        ErrorDialog(_('File not found!')); Exit;
        end;
      end;
    end;
  if not InputQuery(_('Encrypt using public key'),_('User ID:'),PgpID) then Exit;
  sName:=MailFile+'.tmp';
    AssignFile(oMail,sName); rewrite(oMail);  // new output
  // prepare temporary files
  sbe:=AddPath(TempDirectory,'enc.tmp');
  AssignFile(iBlock,sbe); rewrite(iBlock);
  sbd:=AddPath(TempDirectory,'dec.tmp');
  AssignFile(oBlock,sbd);
  if FileExists(sbd) then DeleteFile(sbd);
  if FileExists(sbe) then DeleteFile(sbe);
  StdOut:=TStringList.Create;
  nl:=0;
  for i:=0 to Length(EncIndex)-1 do begin
    StatusBar.SimpleText:=Format(_('Processing section: %u'),[i+1]);
    // copy until next previous boundary
    for j:=nl to EncIndex[i].BlkStart-1 do writeln(oMail,MailText[j]);
    rewrite(iBlock);
    for j:=EncIndex[i].BlkStart to EncIndex[i].BlkEnd do
      writeln(iBlock,MailText[j]);    // encrypted section
    CloseFile(iBlock);
    hr:=ExecuteProcess (Format(GpgDecCommand,[sbe,sbd]),'',StdOut,cpUtf8);  // decrypt
    if (hr<>0) and (hr and UserError=0) then begin
      ErrorDialog(_('Error calling GPG!'));
      Break;
      end;
    n:=Pos(sGpgPlaintext,StdOut.Text)+length(sGpgPlaintext);
    k:=PosEx('[',StdOut.Text,n);
    if k>n then s:=Trim(copy(StdOut.Text,n,k-n-1))
    else s:='';
    if not IsEmptyStr(s) then begin  // retrieve filename
      ReadNxtInt(s,Space,0);  // overread specifier
      ReadNxtInt(s,Space,0);  // overread timestamp
      if not IsEmptyStr(s) then begin   // get filename
        s:=AddPath(TempDirectory,ReplChars(s,'%',Space));
        if FileExists(s) then DeleteFile(s);
        RenameFile(sbd,s);
        sbd:=s;
        end;
      end;
    if FileExists(sbe) then DeleteFile(sbe);
    hr:=ExecuteProcess (Format(GpgEncCommand,[PgpID,sbd,sbe]),'',true,0);  // encrypt again
    if (hr<>0) and (hr and UserError=0) then begin
      ErrorDialog(_('Error calling GPG!'));
      Break;
      end;
    reset(iBlock);
    while not eof(iBlock) do begin
      readln(iBlock,s);
      writeln(oMail,s);
      end;
    CloseFile(iBlock);
    if FileExists(sbd) then DeleteFile(sbd);
    if FileExists(sbe) then DeleteFile(sbe);
    nl:=EncIndex[i].BlkEnd+1;
    end;
  for j:=nl to MailText.Count-1 do writeln(oMail,MailText[j]);
  CloseFile(oMail);
  StdOut.Free;
  StatusBar.SimpleText:=_('Done!');
{$ifndef DEBUG}
  sbd:=MailFile+'.bak';
  if FileExists(sbd) then DeleteFile(sbd);
  RenameFile(MailFile,sbd);
  RenameFile(sName,MailFile);
{$endif}
  end;


end.

